from __future__ import annotations

import logging

import zmq
from progress_api import set_backend

from ..connection import Context
from .logging import ZMQLogHandler
from .progress import ZMQProgressBackend


def init_worker_logging(address: str, level: int = logging.INFO):
    ctx: Context = zmq.Context.instance()
    sock = ctx.socket(zmq.PUSH)
    sock.connect(address)

    h = ZMQLogHandler(sock)
    h.setLevel(level)
    root = logging.getLogger()
    root.setLevel(level)
    root.addHandler(h)

    set_backend(ZMQProgressBackend(sock))



from __future__ import annotations

from logging import Handler, LogRecord

from manylog.connection import Socket
from manylog.messages import LogMsg


class ZMQLogHandler(Handler):
    socket: Socket

    def __init__(self, socket: Socket):
        self.socket = socket

    def handle(self, record: LogRecord) -> LogRecord | bool:  # type: ignore
        if not hasattr(record, "message"):
            record.message = record.msg % record.args
        msg = LogMsg.create(record)
        data = msg.encode()
        self.socket.send(data)
        return record



"""
Types and utilities for ZeroMQ connections.
"""

from __future__ import annotations

from typing import TypeAlias

import zmq

Socket: TypeAlias = zmq.Socket[bytes]
Context: TypeAlias = zmq.Context[Socket]



from progress_api import set_backend
from progress_api.backends.mock import MockProgressBackend

import pytest

from manylog import LogListener


@pytest.fixture
def listener():
    listener = LogListener()
    listener.start()
    try:
        yield listener
    finally:
        listener.close()


@pytest.fixture
def progress_mock():
    mock = MockProgressBackend()
    set_backend(mock)
    try:
        yield mock
    finally:
        set_backend("null")



from __future__ import annotations

import pickle
from dataclasses import dataclass, fields
from enum import Enum
from logging import LogRecord
from uuid import UUID

from msgpack import packb, unpackb  # type: ignore
from typing_extensions import Any, ClassVar, Optional, Self, TypeVar

MC = TypeVar("MC", bound="type[BaseMsg]")
MSG_TYPES = ["LogMsg"]
MSG_IMPLS: dict[MsgType, type[BaseMsg]] = {}


def _message_class(cls: MC) -> MC:
    MSG_IMPLS[cls.type] = cls
    return cls


class MsgType(Enum):
    LOG = 1
    PROGRESS_BEGIN = 10
    PROGRESS_END = 11
    PROGRESS_SET_PARAM = 12
    PROGRESS_SET_METRIC = 13
    PROGRESS_UPDATE = 15


@dataclass
class BaseMsg:
    type: ClassVar[MsgType]
    pid: int
    timestamp: float

    def encode(self) -> bytes:
        return packb((self.type.value, self.to_dict()))

    def to_dict(self) -> dict[str, Any]:
        "Convert the message to a MsgPack-encodable dictionary."
        data: dict[str, Any] = {}
        for f in fields(self):
            val = getattr(self, f.name)
            if isinstance(val, UUID):
                data[f.name] = val.bytes
            else:
                data[f.name] = val

        return data

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        "Reconstruct the message from a MsgPack-encoded dictionary."
        for f in fields(cls):
            if f.type == "UUID":
                data[f.name] = UUID(bytes=data[f.name])

        return cls(**data)


@_message_class
@dataclass
class LogMsg(BaseMsg):
    type = MsgType.LOG

    level: int
    name: str
    message: str
    record: Optional[bytes]

    @classmethod
    def create(cls, rec: LogRecord):
        return LogMsg(
            pid=rec.process or -1,
            timestamp=rec.created,
            level=rec.levelno,
            name=rec.name,
            message=rec.message,
            record=pickle.dumps(rec),
        )

    def decode_log_record(self) -> Optional[LogRecord]:
        if not self.record:
            return None

        obj = pickle.loads(self.record)
        if not isinstance(obj, LogRecord):
            raise TypeError("serialized record not LogRecord")

        return obj


@_message_class
@dataclass
class ProgressBegin(BaseMsg):
    type = MsgType.PROGRESS_BEGIN

    uuid: UUID
    spec: dict[str, Any]


@_message_class
@dataclass
class ProgressEnd(BaseMsg):
    type = MsgType.PROGRESS_END

    uuid: UUID


@_message_class
@dataclass
class ProgressSetParam(BaseMsg):
    type = MsgType.PROGRESS_SET_PARAM

    uuid: UUID
    name: str
    value: int | str | None


@_message_class
@dataclass
class ProgressSetMetric(BaseMsg):
    """
    Progress bar metric messages.
    """

    type = MsgType.PROGRESS_SET_METRIC

    uuid: UUID
    label: str
    metric: int | str | float | None
    format: str | None


@_message_class
@dataclass
class ProgressUpdate(BaseMsg):
    """
    Progress bar update messages.
    """

    type = MsgType.PROGRESS_UPDATE

    uuid: UUID
    incr: int
    state: Optional[str]
    src_state: Optional[str]
    metric: int | str | float | None


def decode_message(data: bytes) -> BaseMsg:
    """
    Decode a logging message from bytes.
    """

    rec = unpackb(data)
    type, attrs = rec
    type = MsgType(type)

    cls = MSG_IMPLS[type]
    return cls.from_dict(attrs)



"""
Listener to obtain and re-inject log messages and progress updates.
"""

from __future__ import annotations

import logging
import os
import warnings
from tempfile import TemporaryDirectory
from threading import Thread
from typing import Optional
from uuid import UUID

import zmq
from progress_api.api import Progress
from progress_api.backends import ProgressBarSpec
from progress_api.config import get_backend

import manylog.connection as x
import manylog.messages as m

_log = logging.getLogger(__name__)


class LogListener:
    address: str | None = None
    context: x.Context
    thread: ListenThread | None = None
    _tmpdir: TemporaryDirectory[str]

    def __init__(self, ctx: Optional[x.Context] = None):
        if ctx is None:
            self.context = zmq.Context.instance()
        else:
            self.context = ctx

    def start(self):
        rt_dir = os.environ.get("XDG_RUNTIME_DIR", None)
        self._tmpdir = TemporaryDirectory(prefix="manylog-", dir=rt_dir)
        try:
            socket = self.context.socket(zmq.PULL, zmq.Socket)
            self.address = f"ipc://{self._tmpdir.name}/logging.ipc"
            socket.bind(self.address)
            self.thread = ListenThread(socket)
            self.thread.start()
        except Exception as e:
            self._tmpdir.cleanup()
            raise e

    def close(self):
        if not self.thread:
            warnings.warn("listener thread not running")
            return
        self.thread.shutdown()
        self.thread = None
        self._tmpdir.cleanup()


class ListenThread(Thread):
    socket: x.Socket
    _shutdown_wanted: bool = False
    active_pbs: dict[UUID, Progress]

    def __init__(self, socket: x.Socket):
        super().__init__(name="manylog-listener")
        self.socket = socket
        self.active_pbs = {}

    def run(self):
        while True:
            # we poll for 250ms, to allow shutdown signals
            evt = self.socket.poll(250)
            if evt:
                data = self.socket.recv(zmq.NOBLOCK)
                try:
                    msg = m.decode_message(data)
                except Exception as e:
                    _log.error("error decoding log message: %s", e)
                    continue
                try:
                    self._dispatch_message(msg)
                except Exception as e:
                    _log.error("error dispatching log message: %s", e)
                    continue
            elif self._shutdown_wanted:
                self.socket.close()
                return

    def shutdown(self):
        self._shutdown_wanted = True
        self.join()

    def _dispatch_message(self, msg: m.BaseMsg) -> None:
        match msg:
            case m.LogMsg():
                self._dispatch_log(msg)
            case m.ProgressBegin(uuid=uuid, spec=spec):
                spec["logger"] = logging.getLogger(spec["logger"])
                spec = ProgressBarSpec(**spec)
                self.active_pbs[uuid] = get_backend().create_bar(spec)
            case m.ProgressEnd(uuid=uuid):
                pb = self.active_pbs[uuid]
                pb.finish()
                del self.active_pbs[uuid]
            case m.ProgressSetParam(uuid=uuid, name="label", value=lbl):
                self.active_pbs[uuid].set_label(lbl)  # type: ignore
            case m.ProgressSetParam(uuid=uuid, name="total", value=tot):
                self.active_pbs[uuid].set_total(tot)  # type: ignore
            case m.ProgressSetMetric(uuid=uuid):
                self.active_pbs[uuid].set_metric(msg.label, msg.metric, msg.format)
            case m.ProgressUpdate(uuid=uuid):
                self.active_pbs[uuid].update(msg.incr, msg.state, msg.src_state, msg.metric)
            case _:
                raise TypeError(f"unsupported message type {type(msg)}")

    def _dispatch_log(self, msg: m.LogMsg) -> None:
        logger = logging.getLogger(msg.name)
        rec = msg.decode_log_record()
        if rec is None:
            raise RuntimeError("non-pickled log messages not supported")
        else:
            logger.handle(rec)



"""
Collect log messages and progress updates from multiple processes.
"""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("manylog")
except PackageNotFoundError:
    # package is not installed
    pass

from .listener import LogListener
from .worker import init_worker_logging

__all__ = ["LogListener", "init_worker_logging"]



import logging
import multiprocessing as mp

import pytest

from manylog import LogListener, init_worker_logging


@pytest.fixture
def listener():
    listener = LogListener()
    listener.start()
    try:
        yield listener
    finally:
        listener.close()


def _worker(cstr: str):
    print("child working")
    init_worker_logging(cstr)
    print("worker initialized")
    log = logging.getLogger()
    log.info("test message from child thread")
    print("worker finished")


def test_single_log(listener: LogListener, caplog: pytest.LogCaptureFixture):
    proc = mp.Process(target=_worker, args=[listener.address])
    proc.start()
    proc.join()
    assert proc.exitcode == 0

    assert any([r.message == "test message from child thread" for r in caplog.records])



from __future__ import annotations

import dataclasses
import os
import time
from typing import Optional
from uuid import UUID, uuid4

from progress_api import Progress
from progress_api.backends import ProgressBackend, ProgressBarSpec

import manylog.messages as m
from manylog.connection import Socket


class ZMQProgressBackend(ProgressBackend):
    socket: Socket
    pid: int

    def __init__(self, socket: Socket) -> None:
        super().__init__()
        self.socket = socket
        self.pid = os.getpid()

    def create_bar(self, spec: ProgressBarSpec) -> Progress:
        """
        Create a new progress bar from the given specification.
        """
        id = uuid4()
        sdict = dataclasses.asdict(spec)
        sdict["logger"] = spec.logger.name
        msg = m.ProgressBegin(self.pid, time.time(), id, sdict)
        self.socket.send(msg.encode())
        return ZMQProgress(self.socket, self.pid, id)


class ZMQProgress(Progress):
    socket: Socket
    pid: int
    uuid: UUID

    def __init__(self, socket: Socket, pid: int, id: UUID):
        self.socket = socket
        self.pid = pid
        self.uuid = id

    def set_label(self, label: Optional[str]) -> None:
        msg = m.ProgressSetParam(self.pid, time.time(), self.uuid, "label", label)
        self.socket.send(msg.encode())

    def set_total(self, total: int) -> None:
        msg = m.ProgressSetParam(self.pid, time.time(), self.uuid, "total", total)
        self.socket.send(msg.encode())

    def set_metric(
        self, label: str, value: int | str | float | None, fmt: str | None = None
    ) -> None:
        msg = m.ProgressSetMetric(self.pid, time.time(), self.uuid, label, value, fmt)
        self.socket.send(msg.encode())

    def update(
        self,
        n: int = 1,
        state: Optional[str] = None,
        src_state: Optional[str] = None,
        metric: int | str | float | None = None,
    ) -> None:
        msg = m.ProgressUpdate(self.pid, time.time(), self.uuid, n, state, src_state, metric)
        self.socket.send(msg.encode())

    def finish(self) -> None:
        msg = m.ProgressEnd(self.pid, time.time(), self.uuid)
        self.socket.send(msg.encode())



import logging
import multiprocessing as mp
import time

from progress_api import make_progress
from progress_api.backends.mock import MockProgressBackend

import pytest

from manylog import LogListener, init_worker_logging


def _worker(cstr: str):
    print("child working")
    init_worker_logging(cstr)
    print("worker initialized")
    log = logging.getLogger()
    log.info("starting job")
    prog = make_progress(log, "tasks", 10)
    for _i in range(10):
        time.sleep(0.01)
        prog.update()
    prog.finish()
    log.info("finished job")
    print("worker finished")


def test_single_progress(
    listener: LogListener, caplog: pytest.LogCaptureFixture, progress_mock: MockProgressBackend
):
    proc = mp.Process(target=_worker, args=[listener.address])
    proc.start()
    proc.join()
    assert proc.exitcode == 0

    assert any([r.message == "starting job" for r in caplog.records])
    assert any([r.message == "finished job" for r in caplog.records])
    for rec in progress_mock.record:
        print("-", rec)
    assert "start tasks 10" in progress_mock.record
    assert "finish tasks" in progress_mock.record
